@@ -85,6 +85,7 @@ import           Data.Functor.Identity (Identity (..))
 import           Data.HashMap.Strict (HashMap)
 import           Data.HashSet (HashSet)
 import           Data.Hashable (Hashable)
+import           Data.Int
 import           Data.IntMap (IntMap)
 import qualified Data.IntMap.Strict as IntMap
 import           Data.IntSet (IntSet)
@@ -109,6 +110,7 @@ import qualified Data.Text.Array as TA
 import qualified Data.Text.Foreign as T
 import qualified Data.Text.Internal as T
 import qualified Data.Time as Time
+import qualified Data.Time.Clock.TAI as Time
 import           Data.Typeable (Typeable)
 import qualified Data.Vector as V
 import qualified Data.Vector.Mutable as MV
@@ -120,13 +122,6 @@ import           Foreign.C.Types ()
 import           Foreign.Ptr (plusPtr, minusPtr)
 import           Foreign.Storable (Storable, sizeOf)
 import           GHC.Generics (Generic)
-#ifdef INTEGER_GMP
-import qualified GHC.Integer.GMP.Internals as I
-import           GHC.Types (Int (I#))
-#else
-import           GHC.Types (Word (W#))
-import qualified GHC.Integer.Simple.Internals as I
-#endif
 import           GHC.Real (Ratio(..))
 import           GHC.TypeLits
 import           Instances.TH.Lift ()
@@ -135,9 +130,29 @@ import           Language.Haskell.TH.Instances ()
 import           Language.Haskell.TH.ReifyMany
 import           Language.Haskell.TH.Syntax
 import           Network.Socket (AddrInfo)
+import           Numeric.Natural (Natural)
 import           Prelude
 import           TH.Derive
 
+#if MIN_VERSION_time(1,8,0)
+import qualified Data.Time.Clock.System as Time
+#endif
+#if MIN_VERSION_time(1,9,0)
+import qualified Data.Time.Format.ISO8601 as Time
+#endif
+#if MIN_VERSION_time(1,11,0)
+import qualified Data.Time.Calendar.Quarter as Time
+import qualified Data.Time.Calendar.WeekDate as Time
+#endif
+
+#ifdef INTEGER_GMP
+import qualified GHC.Integer.GMP.Internals as I
+import           GHC.Types (Int (I#))
+#else
+import           GHC.Types (Word (W#))
+import qualified GHC.Integer.Simple.Internals as I
+#endif
+
 -- Conditional import to avoid warning
 #ifdef INTEGER_GMP
 #if MIN_VERSION_integer_gmp(1,0,0)
@@ -145,6 +160,16 @@ import           GHC.Prim (sizeofByteArray#)
 #endif
 #endif
 
+-- TODO: higher arities?  Limited now by Generics instances for tuples
+$(return $ map deriveTupleStoreInstance [2..7])
+
+$(deriveManyStoreFromStorable
+  (\ty ->
+    case ty of
+      ConT n | elem n [''Char, ''Int, ''Int64, ''Word, ''Word8, ''Word32] -> True
+      _ -> False
+    ))
+
 ------------------------------------------------------------------------
 -- Utilities for defining list-like 'Store' instances in terms of 'IsSequence'
 
@@ -300,25 +325,6 @@ peekOrdMapWith f = do
     f <$> peek
 {-# INLINE peekOrdMapWith #-}
 
-{-
-------------------------------------------------------------------------
--- Utilities for defining list-like 'Store' instances in terms of Foldable
-
--- | Implement 'size' for a 'Foldable' of 'Store' instances. Note that
--- this assumes the extra 'Foldable' structure is discardable - this
--- only serializes the elements.
-sizeListLikeFoldable :: forall t a. (Foldable t, Store a) => Size (t a)
-sizeListLikeFoldable = VarSize $ \t ->
-    case size :: Size e of
-        ConstSize n ->  n * length x + sizeOf (undefined :: Int)
-        VarSize f -> foldl' (\acc x -> acc + f x) (sizeOf (undefined :: Int))
-{-# INLINE sizeSequence #-}
-
-pokeListLikeFoldable :: forall t a. Foldable t => t a -> Poke ()
-pokeListLikeFoldable x = do
-    poke (length x)
--}
-
 ------------------------------------------------------------------------
 -- Utilities for implementing 'Store' instances for list-like mutable things
 
@@ -429,18 +435,27 @@ instance Store SBS.ShortByteString where
         return (SBS.SBS array)
 
 instance Store LBS.ByteString where
-    -- FIXME: faster conversion? Is this ever going to be a problem?
-    --
-    -- I think on 64 bit systems, Int will have 64 bits. On 32 bit
-    -- systems, we'll never exceed the range of Int by this conversion.
     size = VarSize $ \x ->
          sizeOf (undefined :: Int)  +
          fromIntegral (LBS.length x)
-    -- FIXME: more efficient implementation that avoids the double copy
+    -- TODO: more efficient implementation that avoids the double copy
     poke = poke . LBS.toStrict
     peek = fmap LBS.fromStrict peek
 
 instance Store T.Text where
+#if MIN_VERSION_text(2,0,0)
+    size = VarSize $ \x ->
+        sizeOf (undefined :: Int) +
+        T.lengthWord8 x
+    poke x = do
+        let !(T.Text (TA.ByteArray array) w8Off w8Len) = x
+        poke w8Len
+        pokeFromByteArray array w8Off w8Len
+    peek = do
+        w8Len <- peek
+        ByteArray array <- peekToByteArray "Data.Text.Text" w8Len
+        return (T.Text (TA.ByteArray array) 0 w8Len)
+#else
     size = VarSize $ \x ->
         sizeOf (undefined :: Int) +
         2 * (T.lengthWord16 x)
@@ -452,13 +467,11 @@ instance Store T.Text where
         w16Len <- peek
         ByteArray array <- peekToByteArray "Data.Text.Text" (2 * w16Len)
         return (T.Text (TA.Array array) 0 w16Len)
+#endif
 
 ------------------------------------------------------------------------
 -- Known size instances
 
--- TODO: this doesn't scale nicely to 'Text'. Force it to be byte size?
--- 'StaticByteSize'?
-
 newtype StaticSize (n :: Nat) a = StaticSize { unStaticSize :: a }
     deriving (Eq, Show, Ord, Data, Typeable, Generic)
 
@@ -481,7 +494,6 @@ instance KnownNat n => IsStaticSize n BS.ByteString where
 instance KnownNat n => Store (StaticSize n BS.ByteString) where
     size = ConstSize (fromInteger (natVal (Proxy :: Proxy n)))
     poke (StaticSize x) = do
-        -- TODO: worth it to put an assert here?
         let (sourceFp, sourceOffset, sourceLength) = BS.toForeignPtr x
         pokeFromForeignPtr sourceFp sourceOffset sourceLength
     peek = do
@@ -497,7 +509,11 @@ liftStaticSize tyq (StaticSize x) = do
     let numTy = litT $ numTyLit $ natVal (Proxy :: Proxy n)
     [| StaticSize $(lift x) :: StaticSize $(numTy) $(tyq) |]
 
+#if MIN_VERSION_template_haskell(2,17,0)
+staticByteStringExp :: Quote m => BS.ByteString -> m Exp
+#else
 staticByteStringExp :: BS.ByteString -> ExpQ
+#endif
 staticByteStringExp bs =
     [| StaticSize bs :: StaticSize $(litT (numTyLit (fromIntegral len))) BS.ByteString |]
   where
@@ -563,13 +579,11 @@ instance (Eq a, Hashable a, Store a) => Store (HashSet a) where
     peek = peekSet
 
 instance (A.Ix i, Store i, Store e) => Store (A.Array i e) where
-    -- TODO: Speed up poke and peek
     size = sizeArray
     poke = pokeArray
     peek = peekArray
 
 instance (A.Ix i, A.IArray A.UArray e, Store i, Store e) => Store (A.UArray i e) where
-    -- TODO: Speed up poke and peek
     size = sizeArray
     poke = pokeArray
     peek = peekArray
@@ -712,14 +726,16 @@ numDigits = go 0
         go !acc (I.Some _ ds) = go (acc + 1) ds
 #endif
 
--- instance Store GHC.Fingerprint.Types.Fingerprint where
+-- Piggybacks off of the Integer instance
 
-instance Store (Fixed a) where
-    size = contramap (\(MkFixed x) -> x) (size :: Size Integer)
-    poke (MkFixed x) = poke x
-    peek = MkFixed <$> peek
-
--- instance Store a => Store (Tree a) where
+instance Store Natural where
+  size = contramap fromIntegral (size :: Size Integer)
+  poke = poke . toInteger
+  peek = do
+      x <- peek :: Peek Integer
+      if x < 0
+          then peekException "Encountered negative integer when expecting a Natural"
+          else return $ fromIntegral x
 
 ------------------------------------------------------------------------
 -- Other instances
@@ -740,30 +756,17 @@ instance Store a => Store (Ratio a) where
 -- Similarly, manual implementation due to no Generic instance for
 -- Complex and Identity in GHC-7.10 and earlier.
 
-instance Store a => Store (Complex a) where
-    size = combineSize (\(x :+ _) -> x) (\(_ :+ y) -> y)
-    poke (x :+ y) = poke (x, y)
-    peek = uncurry (:+) <$> peek
-
-instance Store a => Store (Identity a) where
-    size = contramap (\(Identity x) -> x) size
-    poke (Identity x) = poke x
-    peek = Identity <$> peek
-
-instance Store Time.Day where
-    size = contramap Time.toModifiedJulianDay (size :: Size Integer)
-    poke = poke . Time.toModifiedJulianDay
-    peek = Time.ModifiedJulianDay <$> peek
+$($(derive [d| instance Deriving (Store (Fixed a)) |]))
 
 instance Store Time.DiffTime where
-    size = contramap (realToFrac :: Time.DiffTime -> Pico) (size :: Size Pico)
-    poke = (poke :: Pico -> Poke ()) . realToFrac
-    peek = Time.picosecondsToDiffTime <$> peek
+    size = contramap (realToFrac :: Time.DiffTime -> Pico) size
+    poke = poke . (realToFrac :: Time.DiffTime -> Pico)
+    peek = (realToFrac :: Pico -> Time.DiffTime) <$> peek
 
-instance Store Time.UTCTime where
-    size = combineSize Time.utctDay Time.utctDayTime
-    poke (Time.UTCTime day time) = poke (day, time)
-    peek = uncurry Time.UTCTime <$> peek
+instance Store Time.NominalDiffTime where
+    size = contramap (realToFrac :: Time.NominalDiffTime -> Pico) size
+    poke = poke . (realToFrac :: Time.NominalDiffTime -> Pico)
+    peek = (realToFrac :: Pico -> Time.NominalDiffTime) <$> peek
 
 instance Store ()
 instance Store a => Store (Dual a)
@@ -774,23 +777,49 @@ instance Store a => Store (Last a)
 instance Store a => Store (Maybe a)
 instance Store a => Store (Const a b)
 
--- FIXME: have TH deriving handle unboxed fields?
-
 ------------------------------------------------------------------------
 -- Instances generated by TH
 
 $($(derive [d|
-    -- TODO
-    -- instance Deriving (Store ())
+    instance Store a => Deriving (Store (Complex a))
+    instance Store a => Deriving (Store (Identity a))
+
     instance Deriving (Store All)
     instance Deriving (Store Any)
     instance Deriving (Store Void)
     instance Deriving (Store Bool)
     instance (Store a, Store b) => Deriving (Store (Either a b))
+
+    instance Deriving (Store Time.AbsoluteTime)
+    instance Deriving (Store Time.Day)
+    instance Deriving (Store Time.LocalTime)
+    instance Deriving (Store Time.TimeOfDay)
+    instance Deriving (Store Time.TimeZone)
+    instance Deriving (Store Time.UTCTime)
+    instance Deriving (Store Time.UniversalTime)
+    instance Deriving (Store Time.ZonedTime)
+    instance Deriving (Store Time.TimeLocale)
+
+#if MIN_VERSION_time(1,8,0)
+    instance Deriving (Store Time.SystemTime)
+#endif
+
+#if MIN_VERSION_time(1,9,0)
+    instance Deriving (Store Time.CalendarDiffDays)
+    instance Deriving (Store Time.CalendarDiffTime)
+    instance Deriving (Store Time.FormatExtension)
+#endif
+
+#if MIN_VERSION_time(1,11,0)
+    instance Deriving (Store Time.DayOfWeek)
+    instance Deriving (Store Time.FirstWeekType)
+    instance Deriving (Store Time.Quarter)
+    instance Deriving (Store Time.QuarterOfYear)
+#endif
+
     |]))
 
--- TODO: higher arities?  Limited now by Generics instances for tuples
-$(return $ map deriveTupleStoreInstance [2..7])
+$(deriveManyStorePrimVector)
 
 $(deriveManyStoreUnboxVector)
 
@@ -817,10 +846,7 @@ $(deriveManyStoreFromStorable
       _ -> True
     ))
 
-$(deriveManyStorePrimVector)
-
 $(reifyManyWithoutInstances ''Store [''ModName, ''NameSpace, ''PkgName] (const True) >>=
---   mapM (\name -> deriveStore [] (ConT name) .dtCons =<< reifyDataType name))
    mapM (\name -> return (deriveGenericInstance [] (ConT name))))
 
 -- Explicit definition needed because in template-haskell <= 2.9 (GHC
@@ -865,5 +891,4 @@ instance Store NameFlavour where
 #endif
 
 $(reifyManyWithoutInstances ''Store [''Info] (const True) >>=
---   mapM (\name -> deriveStore [] (ConT name) .dtCons =<< reifyDataType name))
-   mapM (\name -> return (deriveGenericInstance [] (ConT name))))
+   mapM deriveGenericInstanceFromName)
