@@ -23,6 +23,7 @@ import qualified Data.ByteString.Lazy as LBS
 import qualified Data.ByteString.Short as SBS
 import           Data.Complex (Complex(..))
 import           Data.Containers (mapFromList, setFromList)
+import           Data.Fixed (Pico)
 import           Data.Generics (listify)
 import           Data.HashMap.Strict (HashMap)
 import           Data.HashSet (HashSet)
@@ -33,9 +34,6 @@ import           Data.IntSet (IntSet)
 import qualified Data.List.NonEmpty as NE
 import           Data.Map (Map)
 import           Data.Monoid
-#if !MIN_VERSION_primitive(0,7,0)
-import           Data.Primitive.Types (Addr)
-#endif
 import           Data.Proxy (Proxy(..))
 import           Data.Sequence (Seq)
 import           Data.Sequences (fromList)
@@ -49,12 +47,12 @@ import           Data.StoreSpec.TH
 import           Data.Text (Text)
 import qualified Data.Text as T
 import qualified Data.Time as Time
+import qualified Data.Time.Clock.TAI as Time
 import           Data.Typeable (Typeable)
 import qualified Data.Vector as V
 import qualified Data.Vector.Primitive as PV
 import qualified Data.Vector.Storable as SV
 import qualified Data.Vector.Unboxed as UV
-import           Data.Void (Void)
 import           Data.Word
 import           Foreign.C.Types
 import           Foreign.Ptr
@@ -62,23 +60,41 @@ import           Foreign.Storable (Storable, sizeOf)
 import           GHC.Fingerprint.Type (Fingerprint(..))
 import           GHC.Generics
 import           GHC.Real (Ratio(..))
+#if MIN_VERSION_base(4,15,0)
+import           GHC.RTS.Flags (IoSubSystem(..))
+#endif
 import           Language.Haskell.TH
-import           Language.Haskell.TH.ReifyMany
 import           Language.Haskell.TH.Syntax
 import           Network.Socket
+import           Numeric.Natural (Natural)
 import           Prelude
-import           System.Clock (TimeSpec)
+import           System.Clock (TimeSpec(..))
 import           System.Posix.Types
 import           Test.Hspec hiding (runIO)
 import           Test.SmallCheck.Series
 import           TH.Utilities (unAppsT)
 
+#if !MIN_VERSION_primitive(0,7,0)
+import           Data.Primitive.Types (Addr)
+#endif
+
+#if MIN_VERSION_time(1,8,0)
+import qualified Data.Time.Clock.System as Time
+#endif
+#if MIN_VERSION_time(1,9,0)
+import qualified Data.Time.Format.ISO8601 as Time
+#endif
+#if MIN_VERSION_time(1,11,0)
+import qualified Data.Time.Calendar.Quarter as Time
+#endif
+
+#if !MIN_VERSION_smallcheck(1,2,0)
+import           Data.Void (Void)
+#endif
+
 ------------------------------------------------------------------------
 -- Instances for base types
 
--- TODO: should be possible to do something clever where it only defines
--- instances that don't already exist.  For now, just doing it manually.
-
 addMinAndMaxBounds :: forall a. (Bounded a, Eq a) => [a] -> [a]
 addMinAndMaxBounds xs =
     (if (minBound :: a) `notElem` xs then [minBound] else []) ++
@@ -89,26 +105,38 @@ $(mkManyHasTypeHash [ [t| Int32 |] ])
 -- Serial instances for (Num a, Bounded a) types. Only really
 -- appropriate for the use here.
 
+#if !MIN_VERSION_network(3,1,2)
 instance Bounded PortNumber where
   minBound = 0
   maxBound = 65535
+#endif
 
-$(do let ns = [ ''CWchar, ''CUShort, ''CULong, ''CULLong, ''CIntMax
+$(do let ns = [ ''PortNumber
+
+#if !MIN_VERSION_smallcheck(1,2,0)
+              , ''CWchar, ''CUShort, ''CULong, ''CULLong, ''CIntMax
               , ''CUIntMax, ''CPtrdiff, ''CSChar, ''CShort, ''CUInt, ''CLLong
-              , ''CLong, ''CInt, ''CChar, ''CSsize, ''CPid
+              , ''CLong, ''CInt, ''CChar
+#endif
+              , ''CSsize, ''CPid
               , ''COff, ''CMode, ''CIno, ''CDev
 #if !MIN_VERSION_smallcheck(1,1,4)
               , ''Word8, ''Word16, ''Word32, ''Word64
               , ''Int8, ''Int16, ''Int32, ''Int64
 #endif
-              , ''PortNumber
 #if !MIN_VERSION_smallcheck(1,1,3)
               , ''Word
 #endif
 #if MIN_VERSION_base(4,10,0)
-              , ''CBool, ''CClockId, ''CKey, ''CId
+#if !MIN_VERSION_smallcheck(1,2,0)
+              , ''CBool
+#endif
+              , ''CClockId, ''CKey, ''CId
               , ''CBlkSize, ''CFsBlkCnt, ''CFsFilCnt, ''CBlkCnt
 #endif
+#if MIN_VERSION_base(4,14,0)
+              , ''CSocklen, ''CNfds
+#endif
 #ifndef mingw32_HOST_OS
               , ''CUid, ''CTcflag, ''CRLim, ''CNlink, ''CGid
 #endif
@@ -121,14 +149,19 @@ $(do let ns = [ ''CWchar, ''CUShort, ''CULong, ''CULLong, ''CIntMax
 -- Serial instances for (Num a) types. Only really appropriate for the
 -- use here.
 
-$(do let ns = [ ''CUSeconds, ''CClock, ''CTime, ''CUChar, ''CSize, ''CSigAtomic
+$(do let ns =
+#if !MIN_VERSION_smallcheck(1,2,0)
+              [ ''CUSeconds, ''CClock, ''CTime, ''CUChar, ''CSize, ''CSigAtomic
               ,  ''CSUSeconds, ''CFloat, ''CDouble
               ] ++
-#ifdef mingw32_HOST_OS
-              []
-#else
-              [ ''CSpeed, ''CCc ]
 #endif
+#if !MIN_VERSION_smallcheck(1,1,3)
+              [ ''Natural ] ++
+#endif
+#ifndef mingw32_HOST_OS
+              [ ''CSpeed, ''CCc ] ++
+#endif
+              []
          f n = [d| instance Monad m => Serial m $(conT n) where
                       series = generate (\_ -> [0, 1]) |]
      concat <$> mapM f ns)
@@ -140,32 +173,6 @@ $(do tys <- getAllInstanceTypes1 ''PV.Prim
                       series = fmap PV.fromList series |]
      concat <$> mapM f (filter (\ty -> length (unAppsT ty) == 1) tys))
 
-{-
--- Instances for TH are not currently needed, because the tests for
--- serialization roundtripping are disabled due to Bytes' Eq instance,
--- see issue #150.
-
--- Needs to be done manually because in GHC 7.8's TH, NameFlavour uses
--- unboxed values and cannot use generic deriving. So we skip having an
--- instance for it.
-instance Monad m => Serial m Name where series = fmap mkName series
-
-instance Monad m => Serial m Bytes where
-  series = fmap ((\(BS.PS p o l) -> Bytes p (fromIntegral o) (fromIntegral l)) . BS.pack) series
-
--- Serial instances for (Generic a) types.
-
--- FIXME: generating for TH instances is probably just adding
--- unnecessary compiletime + runtime overhead.
-$(do thNames <- reifyManyWithoutInstances
-         ''Serial
-         [''Info, ''Loc, ''ModName, ''PkgName, ''NameSpace, ''OccName]
-         (`notElem` [''NameFlavour])
-     let ns = [ ''Any, ''All ] ++ thNames
-         f n = [d| instance Monad m => Serial m $(conT n) |]
-     concat <$> mapM f ns)
--}
-
 $(do let ns = [ ''Dual, ''Sum, ''Product, ''First, ''Last ]
          f n = [d| instance (Monad m, Serial m a) => Serial m ($(conT n) a) |]
      concat <$> mapM f ns)
@@ -212,8 +219,6 @@ instance Monad m => Serial m Text where
 instance (Monad m, Serial m a) => Serial m (Seq a) where
     series = fmap fromList series
 
-instance (Monad m, Serial m a) => Serial m (Complex a) where
-    series = uncurry (:+) <$> (series >< series)
 
 instance (Monad m, Serial m a, UV.Unbox a) => Serial m (UV.Vector a) where
     series = fmap fromList series
@@ -240,16 +245,31 @@ instance Monad m => Serial m Time.Day where
 instance Monad m => Serial m Time.DiffTime where
     series = Time.picosecondsToDiffTime <$> series
 
+instance Monad m => Serial m Time.NominalDiffTime where
+    series = (realToFrac :: Integer -> Time.NominalDiffTime) <$> series
+
 instance Monad m => Serial m Time.UTCTime where
     series = uncurry Time.UTCTime <$> (series >< series)
 
-instance (Monad m, Serial m a) => Serial m (NE.NonEmpty a)
-
 instance (Monad m, Serial m a) => Serial m (Tagged a)
 
--- Should probably get added to smallcheck :)
+#if MIN_VERSION_base(4,15,0)
+instance Monad m => Serial m IoSubSystem where
+  series = cons0 IoPOSIX \/ cons0 IoNative
+#endif
+
+#if !MIN_VERSION_smallcheck(1,2,0)
+instance (Monad m, Serial m a) => Serial m (Complex a) where
+    series = uncurry (:+) <$> (series >< series)
+
+instance (Monad m, Serial m a) => Serial m (NE.NonEmpty a)
+
 instance Monad m => Serial m Void where
     series = generate (\_ -> [])
+#endif
+
+instance Monad m => Serial m TimeSpec where
+    series = uncurry TimeSpec <$> (series >< series)
 
 -- We define our own Serial instance for 'Ratio' because of <https://github.com/feuerbach/smallcheck/pull/34>
 
@@ -311,7 +331,34 @@ spec = do
 #if MIN_VERSION_base(4,10,0)
                  , [t| CTimer |]
 #endif
-                 , [t| TimeSpec |]
+
+-- Assume the TH generated instances for Time work, to avoid defining
+-- Serial instances. Also some lack Show / Eq.
+
+                 , [t| Time.AbsoluteTime |]
+                 , [t| Time.Day |]
+                 , [t| Time.LocalTime |]
+                 , [t| Time.TimeOfDay |]
+                 , [t| Time.TimeZone |]
+                 , [t| Time.UTCTime |]
+                 , [t| Time.UniversalTime |]
+                 , [t| Time.ZonedTime |]
+                 , [t| Time.TimeLocale |]
+#if MIN_VERSION_time(1,8,0)
+                 , [t| Time.SystemTime |]
+#endif
+#if MIN_VERSION_time(1,9,0)
+                 , [t| Time.FormatExtension |]
+                 , [t| Time.CalendarDiffDays |]
+                 , [t| Time.CalendarDiffTime |]
+#endif
+#if MIN_VERSION_time(1,11,0)
+                 , [t| Time.DayOfWeek |]
+                 , [t| Time.FirstWeekType |]
+                 , [t| Time.Quarter |]
+                 , [t| Time.QuarterOfYear |]
+#endif
+
                  ]
              omitTys <- (omitTys0 ++) <$> mapM (\ty -> [t| PV.Vector $(pure ty) |]) omitTys0
              let f ty = isMonoType ty && ty `notElem` omitTys && null (listify isThName ty)
@@ -439,6 +486,10 @@ spec = do
         (decodeEx bs :: HashMap Int ()) `shouldBe` m
         evaluate (decodeEx bs :: Map Int ()) `shouldThrow` isUnexpectedMarkerException
         evaluate (decodeEx bs :: IntMap ()) `shouldThrow` isUnexpectedMarkerException
+    it "Expects decode of negative integer as a natural to throw PeekException" $ do
+        evaluate (decodeEx (encode ((-5) :: Integer)) :: Natural)
+            `shouldThrow` isNegativeNaturalException
+
 
 isPokeException :: Test.Hspec.Selector PokeException
 isPokeException = const True
@@ -452,3 +503,7 @@ isTooManyBytesException (PeekException _ t) = "Attempted to read too many bytes"
 isUnexpectedMarkerException :: Test.Hspec.Selector PeekException
 isUnexpectedMarkerException (PeekException _ t) =
     "Expected marker for ascending Map / IntMap: " `T.isPrefixOf` t
+
+isNegativeNaturalException :: Test.Hspec.Selector PeekException
+isNegativeNaturalException (PeekException _ t) =
+    "Encountered negative integer when expecting a Natural" `T.isPrefixOf` t
